from os import path
from tempfile import TemporaryDirectory
from typing import Optional

import pytest
import yaml

from pytest_helm_templates.helm_runner import HelmRunner
from pytest_helm_templates_test.test_helpers import fixture_path


def test_computed_values_raises_error_if_local_chart_not_found() -> None:
    with pytest.raises(ValueError) as ex:
        HelmRunner().computed_values(chart="/almost/certainly/not/a/real/path")

    expected_error = "Computed values can only be rendered for local charts."
    assert expected_error in str(ex)


@pytest.mark.parametrize(
    "use_relative_chart_path",
    (False, True),
)
def test_computed_values_returns_expected_values(use_relative_chart_path: bool) -> None:
    test_chart_absolute_path = fixture_path("charts/test-chart")
    test_chart_path = test_chart_absolute_path
    charts_path: Optional[str] = None
    if use_relative_chart_path:
        charts_path = fixture_path("charts")
        test_chart_path = path.relpath(test_chart_path, charts_path)

    helm_runner = HelmRunner(cwd=charts_path)
    values = helm_runner.computed_values(chart=test_chart_path)
    with open(
        f"{test_chart_absolute_path}/values.yaml",
        encoding="utf-8",
        mode="r",
    ) as file:
        expected_values = yaml.safe_load(file)
    assert values == expected_values


def test_computed_values_can_handle_values_given_as_a_dict() -> None:
    test_chart_path = fixture_path("charts/test-chart")

    helm_runner = HelmRunner()
    values = helm_runner.computed_values(
        chart=test_chart_path,
        values=[{"serviceAccount": {"create": False}}],
    )
    with open(
        f"{test_chart_path}/values.yaml",
        encoding="utf-8",
        mode="r",
    ) as file:
        expected_values = yaml.safe_load(file)

    # Make expected values match our overridden value
    expected_values["serviceAccount"]["create"] = False
    assert values == expected_values


def test_computed_values_includes_dependency_values() -> None:
    test_chart_path = fixture_path("charts/test-chart")

    helm_runner = HelmRunner()
    values = helm_runner.computed_values(
        chart=test_chart_path,
        values=[{"dependency": {"enabled": True}}],
    )
    with open(
        f"{test_chart_path}/values.yaml",
        encoding="utf-8",
        mode="r",
    ) as file:
        expected_values = yaml.safe_load(file)

    # Make expected values include expected dependency values
    expected_values["dependency"] = {
        "config": True,
        "enabled": True,
        "global": {},
    }
    assert values == expected_values


@pytest.mark.parametrize(
    "use_relative_chart_path",
    (False, True),
)
def test_dependency_build_runs_without_exception(use_relative_chart_path: bool) -> None:
    test_chart_absolute_path = fixture_path("charts/test-chart")
    test_chart_path = test_chart_absolute_path
    charts_path: Optional[str] = None
    if use_relative_chart_path:
        charts_path = fixture_path("charts")
        test_chart_path = path.relpath(test_chart_path, charts_path)

    helm_runner = HelmRunner(cwd=charts_path)
    helm_runner.dependency_build(chart=test_chart_path)


@pytest.mark.parametrize(
    "use_relative_chart_path",
    (False, True),
)
def test_dependency_update_runs_without_exception(
    use_relative_chart_path: bool,
) -> None:
    test_chart_absolute_path = fixture_path("charts/test-chart")
    test_chart_path = test_chart_absolute_path
    charts_path: Optional[str] = None
    if use_relative_chart_path:
        charts_path = fixture_path("charts")
        test_chart_path = path.relpath(test_chart_path, charts_path)

    helm_runner = HelmRunner(cwd=charts_path)
    helm_runner.dependency_update(chart=test_chart_path)


def test_notes_raises_error_if_local_chart_not_found() -> None:
    with pytest.raises(ValueError) as ex:
        HelmRunner().notes(chart="/almost/certainly/not/a/real/path", name="test-chart")

    expected_error = "Notes can only be rendered for local charts."
    assert expected_error in str(ex)


def test_notes_raises_error_if_notes_txt_not_found() -> None:
    with TemporaryDirectory() as temporary_directory_path:
        with pytest.raises(ValueError) as ex:
            HelmRunner().notes(chart=temporary_directory_path, name="test-chart")

    expected_error = (
        "Unable to find notes template at"
        f" `{temporary_directory_path}/templates/NOTES.txt`"
    )
    assert expected_error in str(ex)


@pytest.mark.parametrize(
    "use_relative_chart_path",
    (False, True),
)
def test_notes_returns_expected_notes(use_relative_chart_path: bool) -> None:
    test_chart_path = fixture_path("charts/test-chart")
    charts_path: Optional[str] = None
    if use_relative_chart_path:
        charts_path = fixture_path("charts")
        test_chart_path = path.relpath(test_chart_path, charts_path)

    helm_runner = HelmRunner(cwd=charts_path)
    notes = helm_runner.notes(chart=test_chart_path, name="test-chart")
    expected_notes_excerpt = "Visit http://127.0.0.1:8080 to use your application"
    assert expected_notes_excerpt in notes


def test_notes_can_handle_values_given_a_dict() -> None:
    test_chart_path = fixture_path("charts/test-chart")

    helm_runner = HelmRunner()
    notes = helm_runner.notes(
        chart=test_chart_path,
        name="test-chart",
        values=[{"service": {"type": "LoadBalancer"}}],
    )
    expected_notes_excerpt = "It may take a few minutes for the LoadBalancer IP"
    assert expected_notes_excerpt in notes


@pytest.mark.parametrize(
    "use_relative_chart_path",
    (False, True),
)
def test_template_returns_expected_helm_template_output(
    use_relative_chart_path: bool,
) -> None:
    test_chart_path = fixture_path("charts/test-chart")
    charts_path: Optional[str] = None
    if use_relative_chart_path:
        charts_path = fixture_path("charts")
        test_chart_path = path.relpath(test_chart_path, charts_path)

    helm_runner = HelmRunner(cwd=charts_path)
    manifests = helm_runner.template(chart=test_chart_path, name="test-chart")

    manifest_names = {manifest["metadata"]["name"] for manifest in manifests}
    expected_manifest_names = {
        "test-chart-deployment",
        "test-chart-service",
        "test-chart-service-account",
        "test-chart-test-connection",
    }
    assert expected_manifest_names == manifest_names


def test_template_returns_expected_helm_template_output_for_remote_chart() -> None:
    helm_runner = HelmRunner()
    manifests = helm_runner.template(
        chart="hello-world",
        name="test-chart",
        repo="https://helm.github.io/examples",
        version="0.1.0",
    )

    manifest_names = {manifest["metadata"]["name"] for manifest in manifests}
    expected_manifest_names = {"test-chart-hello-world"}
    assert expected_manifest_names == manifest_names

    manifest_kinds = {manifest["kind"] for manifest in manifests}
    expected_manifest_kinds = {"Deployment", "Service", "ServiceAccount"}
    assert expected_manifest_kinds == manifest_kinds


def test_template_can_handle_values_given_a_dict() -> None:
    test_chart_path = fixture_path("charts/test-chart")

    helm_runner = HelmRunner()
    manifests = helm_runner.template(
        chart=test_chart_path,
        name="test-chart",
        values=[{"serviceAccount": {"create": False}}],
    )

    manifest_names = {manifest["metadata"]["name"] for manifest in manifests}
    assert "test-chart-service-account" not in manifest_names


@pytest.mark.parametrize(
    "use_relative_chart_path",
    (False, True),
)
def test_values_returns_expected_values(use_relative_chart_path: bool) -> None:
    test_chart_absolute_path = fixture_path("charts/test-chart")
    test_chart_path = test_chart_absolute_path
    charts_path: Optional[str] = None
    if use_relative_chart_path:
        charts_path = fixture_path("charts")
        test_chart_path = path.relpath(test_chart_path, charts_path)

    helm_runner = HelmRunner(cwd=charts_path)
    values = helm_runner.values(chart=test_chart_path)
    with open(
        f"{test_chart_absolute_path}/values.yaml",
        encoding="utf-8",
        mode="r",
    ) as file:
        expected_values = yaml.safe_load(file)
    assert values == expected_values


def test_values_returns_expected_values_for_remote_chart() -> None:
    helm_runner = HelmRunner()
    values = helm_runner.values(
        chart="hello-world",
        repo="https://helm.github.io/examples",
        version="0.1.0",
    )
    assert "image" in values
    assert "replicaCount" in values



from typing import List, Optional


class ShowValuesCommand:
    @classmethod
    def helm_arguments(
        cls,
        chart: str,
        repo: Optional[str] = None,
        version: Optional[str] = None,
    ) -> List[str]:
        components = ["helm", "show", "values", chart]

        if repo:
            components.append("--repo")
            components.append(repo)

        if version:
            components.append("--version")
            components.append(version)

        return components



from os import path
from pathlib import Path


def fixture_path(fixture_path: str) -> str:
    fixtures_path = Path(__file__).parent
    fixture_file_path = fixtures_path.joinpath(fixture_path)
    if not path.exists(fixture_file_path):
        raise ValueError(f"Could not find fixture file: '{fixture_path}'")

    return str(fixture_file_path.absolute())


def read_fixture_file(fixture_file_path: str) -> str:
    with open(fixture_path(fixture_file_path), encoding="utf-8", mode="r") as file:
        return file.read()



from pytest_helm_templates.commands.show_values_command import ShowValuesCommand
from pytest_helm_templates.commands.template_command import TemplateCommand


__all__ = [
    "ShowValuesCommand",
    "TemplateCommand",
]



from typing import List, Optional


class TemplateCommand:
    @classmethod
    def helm_arguments(
        cls,
        chart: str,
        name: str,
        api_versions: Optional[List[str]] = None,
        dry_run: Optional[str] = None,
        include_crds: Optional[bool] = None,
        is_upgrade: Optional[bool] = None,
        kube_version: Optional[str] = None,
        namespace: Optional[str] = None,
        no_hooks: Optional[bool] = None,
        repo: Optional[str] = None,
        show_only: Optional[List[str]] = None,
        skip_tests: Optional[bool] = None,
        values: Optional[List[str]] = None,
        version: Optional[str] = None,
    ) -> List[str]:
        _dry_run = dry_run or "client"
        _include_crds = include_crds if include_crds is not None else True
        _is_upgrade = is_upgrade or False
        _no_hooks = no_hooks or False
        _skip_tests = skip_tests or False
        _values = values if values is not None else []

        components = ["helm", "template", name, chart]

        if api_versions:
            comma_separated_api_versions = ",".join(api_versions)
            components.append("--api-versions")
            components.append(comma_separated_api_versions)

        components.append(f"--dry-run={_dry_run}")

        if _include_crds:
            components.append("--include-crds")

        if _is_upgrade:
            components.append("--is-upgrade")

        if kube_version:
            components.append("--kube-version")
            components.append(kube_version)

        if namespace:
            components.append("--namespace")
            components.append(namespace)

        if _no_hooks:
            components.append("--no-hooks")

        if repo:
            components.append("--repo")
            components.append(repo)

        if show_only:
            for template_path in show_only:
                components.append("--show-only")
                components.append(template_path)

        if _skip_tests:
            components.append("--skip-tests")

        if _values:
            for values_file_or_url in _values:
                components.append("--values")
                components.append(values_file_or_url)

        if version:
            components.append("--version")
            components.append(version)

        return components



import subprocess
import textwrap
from os import path
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import IO, Any, Dict, List, Optional, Tuple, Union
from uuid import uuid4

import yaml

from pytest_helm_templates.commands import ShowValuesCommand, TemplateCommand


class HelmRunner:
    def __init__(
        self,
        cwd: Optional[str] = None,
        env: Optional[Dict[str, str]] = None,
    ) -> None:
        self.cwd = cwd
        self.env = env

    def values(
        self,
        chart: str,
        repo: Optional[str] = None,
        version: Optional[str] = None,
    ) -> Dict:
        """
        Collect the values of the given chart.
        """
        helm_arguments = ShowValuesCommand.helm_arguments(
            chart=chart,
            repo=repo,
            version=version,
        )

        values_output = self._run(helm_arguments=helm_arguments)
        values = yaml.safe_load(values_output)
        if not isinstance(values, Dict):
            raise ValueError(
                "Unexpected values. Expected dict, got" f" {type(values)}: {values}"
            )
        return values

    def computed_values(
        self,
        chart: str,
        values: Optional[List[Union[Dict[str, Any], str]]] = None,
    ) -> Dict:
        """
        Collect the whole tree of values from the given chart and its
        dependencies. This function is a bit hacky in that it depends on having
        a local chart to inject an adhoc template into. There are ways to work
        around that, but an adhoc template seems simpler and less error prone.
        """
        chart_path = Path(chart) if not self.cwd else Path(self.cwd).joinpath(chart)
        if not path.exists(chart_path):
            # We could instead take COMPUTED VALUES from `helm install [name]
            # [chart] --dry-run --debug -f <your_values_file>`, but for now
            # I've taken the easy way out.
            # https://github.com/helm/helm/issues/6772
            raise ValueError(
                "Computed values can only be rendered for local charts. Could"
                f" not find local chart `{chart}` ({str(chart_path)})"
            )

        templates_dir_path = chart_path.joinpath("templates")

        with NamedTemporaryFile(
            dir=templates_dir_path,
            encoding="utf-8",
            mode="w",
        ) as temp_file:
            temp_file_name = path.basename(temp_file.name)
            temp_file.write("{{ toYaml .Values }}")
            temp_file.flush()
            manifests = self.template(
                chart=chart,
                dry_run="client",
                include_crds=False,
                name=str(uuid4()),
                show_only=[f"templates/{temp_file_name}"],
                skip_tests=True,
                values=values,
            )
            values_output = manifests[0]
            if not isinstance(values_output, Dict):
                raise ValueError(
                    "Unexpected computed values. Expected dict, got"
                    f" {type(values_output)}: {values_output}"
                )
            return values_output

    def dependency_build(self, chart: str) -> None:
        self._run(
            [
                "helm",
                "dependency",
                "build",
                chart,
            ]
        )

    def dependency_update(self, chart: str) -> None:
        self._run(
            [
                "helm",
                "dependency",
                "update",
                chart,
            ]
        )

    def notes(
        self,
        chart: str,
        name: str,
        api_versions: Optional[List[str]] = None,
        dry_run: Optional[str] = None,
        is_upgrade: Optional[bool] = None,
        kube_version: Optional[str] = None,
        namespace: Optional[str] = None,
        repo: Optional[str] = None,
        values: Optional[List[Union[Dict[str, Any], str]]] = None,
        version: Optional[str] = None,
    ) -> str:
        """
        This function is a bit hacky in that it requires creating a copy of
        NOTES.txt that's embedded in a YAML template, as such, it depends on a
        local chart. `helm template` doesn't currently return NOTES.txt, and
        otherwise NOTES.txt is only rendered during `helm install`.

        Related helm issue: https://github.com/helm/helm/issues/6901
        """
        chart_path = Path(chart) if not self.cwd else Path(self.cwd).joinpath(chart)
        if not path.exists(chart_path):
            raise ValueError(
                "Notes can only be rendered for local charts. Could not find local"
                f" chart `{chart}` ({str(chart_path)})"
            )

        templates_dir_path = chart_path.joinpath("templates")
        notes_path = templates_dir_path.joinpath("NOTES.txt")
        if not path.exists(notes_path):
            raise ValueError(f"Unable to find notes template at `{notes_path}`")

        with open(notes_path, encoding="utf-8", mode="r") as notes_file:
            notes_template = notes_file.read()
        indented_notes_template = textwrap.indent(notes_template, "  ")

        with NamedTemporaryFile(
            dir=templates_dir_path,
            encoding="utf-8",
            mode="w",
        ) as temp_file:
            temp_file_name = path.basename(temp_file.name)
            temp_file.write(f"---\nNOTES.txt: |\n{indented_notes_template}")
            temp_file.flush()
            manifests = self.template(
                api_versions=api_versions,
                chart=chart,
                dry_run=dry_run,
                include_crds=False,
                is_upgrade=is_upgrade,
                kube_version=kube_version,
                namespace=namespace,
                name=name,
                repo=repo,
                show_only=[f"templates/{temp_file_name}"],
                skip_tests=True,
                values=values,
                version=version,
            )
            notes_output = manifests[0]["NOTES.txt"]
            if not isinstance(notes_output, str):
                raise ValueError(
                    "Unexpected notes template output. Expected string, got"
                    " {type(notes_output)}: {notes_output}"
                )
            return notes_output

    def template(
        self,
        chart: str,
        name: str,
        api_versions: Optional[List[str]] = None,
        dry_run: Optional[str] = None,
        include_crds: Optional[bool] = None,
        is_upgrade: Optional[bool] = None,
        kube_version: Optional[str] = None,
        namespace: Optional[str] = None,
        repo: Optional[str] = None,
        show_only: Optional[List[str]] = None,
        skip_tests: Optional[bool] = None,
        values: Optional[List[Union[Dict[str, Any], str]]] = None,
        version: Optional[str] = None,
    ) -> List[Dict]:
        chart_path = Path(chart) if not self.cwd else Path(self.cwd).joinpath(chart)

        _values = []
        temp_files: List[IO] = []
        try:
            if values:
                for values_instance in values:
                    if isinstance(values_instance, str):
                        _values.append(values_instance)
                    else:
                        temp_file_path, temp_file = self._reify_values(values_instance)
                        _values.append(temp_file_path)
                        temp_files.append(temp_file)
            helm_arguments = TemplateCommand.helm_arguments(
                api_versions=api_versions,
                chart=str(chart_path),
                dry_run=dry_run,
                include_crds=include_crds,
                is_upgrade=is_upgrade,
                kube_version=kube_version,
                namespace=namespace,
                name=name,
                repo=repo,
                show_only=show_only,
                skip_tests=skip_tests,
                values=_values,
                version=version,
            )
            templates_yaml = self._run(helm_arguments)
        finally:
            for temp_file in temp_files:
                temp_file.close()

        return list(yaml.safe_load_all(templates_yaml))

    def _reify_values(self, values: Dict) -> Tuple[str, IO]:
        temp_file = NamedTemporaryFile(delete=False, mode="w")
        temp_file.write(yaml.safe_dump(values))
        temp_file.flush()
        return temp_file.name, temp_file

    def _run(self, helm_arguments: List[str]) -> str:
        completed_process = subprocess.run(
            helm_arguments,
            capture_output=True,
            cwd=self.cwd,
            env=self.env,
        )

        return_code = completed_process.returncode
        if return_code > 0:
            stderr = completed_process.stderr.decode("utf-8")
            raise RuntimeError(
                f"helm command failed with return code {return_code}:"
                f"exec {helm_arguments}\n{stderr}"
            )

        return completed_process.stdout.decode("utf-8")



from typing import List

import pytest

from pytest_helm_templates.commands.template_command import TemplateCommand


@pytest.mark.parametrize(
    "actual_arguments,expected_arguments",
    (
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                name="name",
            ),
            ["helm", "template", "name", "chart", "--dry-run=client", "--include-crds"],
        ),
        (
            TemplateCommand.helm_arguments(
                api_versions=[],
                chart="chart",
                name="name",
            ),
            ["helm", "template", "name", "chart", "--dry-run=client", "--include-crds"],
        ),
        (
            TemplateCommand.helm_arguments(
                api_versions=["v1", "apps/v1"],
                chart="chart",
                name="name",
            ),
            [
                "helm",
                "template",
                "name",
                "chart",
                "--api-versions",
                "v1,apps/v1",
                "--dry-run=client",
                "--include-crds",
            ],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                dry_run="server",
                name="name",
            ),
            ["helm", "template", "name", "chart", "--dry-run=server", "--include-crds"],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                include_crds=False,
                name="name",
            ),
            ["helm", "template", "name", "chart", "--dry-run=client"],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                is_upgrade=False,
                name="name",
            ),
            ["helm", "template", "name", "chart", "--dry-run=client", "--include-crds"],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                is_upgrade=True,
                name="name",
            ),
            [
                "helm",
                "template",
                "name",
                "chart",
                "--dry-run=client",
                "--include-crds",
                "--is-upgrade",
            ],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                kube_version="v1.29.0",
                name="name",
            ),
            [
                "helm",
                "template",
                "name",
                "chart",
                "--dry-run=client",
                "--include-crds",
                "--kube-version",
                "v1.29.0",
            ],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                name="name",
                namespace="test-namespace",
            ),
            [
                "helm",
                "template",
                "name",
                "chart",
                "--dry-run=client",
                "--include-crds",
                "--namespace",
                "test-namespace",
            ],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                name="name",
                no_hooks=False,
            ),
            ["helm", "template", "name", "chart", "--dry-run=client", "--include-crds"],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                name="name",
                no_hooks=True,
            ),
            [
                "helm",
                "template",
                "name",
                "chart",
                "--dry-run=client",
                "--include-crds",
                "--no-hooks",
            ],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                name="name",
                repo="test-repo",
            ),
            [
                "helm",
                "template",
                "name",
                "chart",
                "--dry-run=client",
                "--include-crds",
                "--repo",
                "test-repo",
            ],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                name="name",
                show_only=[],
            ),
            ["helm", "template", "name", "chart", "--dry-run=client", "--include-crds"],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                name="name",
                show_only=["templates/test1.yaml", "templates/test2.yaml"],
            ),
            [
                "helm",
                "template",
                "name",
                "chart",
                "--dry-run=client",
                "--include-crds",
                "--show-only",
                "templates/test1.yaml",
                "--show-only",
                "templates/test2.yaml",
            ],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                name="name",
                skip_tests=True,
            ),
            [
                "helm",
                "template",
                "name",
                "chart",
                "--dry-run=client",
                "--include-crds",
                "--skip-tests",
            ],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                name="name",
                values=[],
            ),
            ["helm", "template", "name", "chart", "--dry-run=client", "--include-crds"],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                name="name",
                values=["values1.yaml", "values2.yaml"],
            ),
            [
                "helm",
                "template",
                "name",
                "chart",
                "--dry-run=client",
                "--include-crds",
                "--values",
                "values1.yaml",
                "--values",
                "values2.yaml",
            ],
        ),
        (
            TemplateCommand.helm_arguments(
                chart="chart",
                name="name",
                version="1.1.1",
            ),
            [
                "helm",
                "template",
                "name",
                "chart",
                "--dry-run=client",
                "--include-crds",
                "--version",
                "1.1.1",
            ],
        ),
    ),
)
def test_helm_arguments_yields_expected_arguments(
    actual_arguments: List[str],
    expected_arguments: List[str],
) -> None:
    assert expected_arguments == actual_arguments



from typing import List

import pytest

from pytest_helm_templates.commands.show_values_command import ShowValuesCommand


@pytest.mark.parametrize(
    "actual_arguments,expected_arguments",
    (
        (
            ShowValuesCommand.helm_arguments(chart="chart"),
            ["helm", "show", "values", "chart"],
        ),
        (
            ShowValuesCommand.helm_arguments(chart="chart", repo="test-repo"),
            ["helm", "show", "values", "chart", "--repo", "test-repo"],
        ),
        (
            ShowValuesCommand.helm_arguments(chart="chart", version="1.1.1"),
            ["helm", "show", "values", "chart", "--version", "1.1.1"],
        ),
    ),
)
def test_helm_arguments_yields_expected_arguments(
    actual_arguments: List[str],
    expected_arguments: List[str],
) -> None:
    assert expected_arguments == actual_arguments
